#include "ObjectStreamer.h"

#include "Config.h"
#include "Debug.h"
#include "LevelManager.h"
#include "MathUtils.h"
#include "ScrollManager.h"

#include <KDebug.h>


// UNSPAWNED/SPAWN only indicates that the object has been created through the
// object streamer. Once an object is SPAWNed, then it's slot is active until
// the object manager has to destroy it when it goes offscreen. It doesn't 
// indicate if the object has been destroyed by the player. An object destroyed
// by the player will be SPAWNED.
#define UNSPAWNED   0 
#define SPAWNED     1

void (*ObjectStreamer_Update)();

u8 streamedObjectStates[256];

s16 currentSectionIndex = 0;
s16 oldCurrentSectionIndex = 0;
u16 numberOfSectionsInAScreen = 0;

void SpawnSection(SpawnList spawnList);

void ObjectStreamer_HorizontalInit();
void ObjectStreamer_VerticalInit();

void Update();
void ObjectStreamer_HorizontalUpdate();
void ObjectStreamer_VerticalUpdate();

void ObjectStreamer_Init()
{
    memset(streamedObjectStates, UNSPAWNED, sizeof(streamedObjectStates));

    if (currentForegroundMap->mapType == MAPTYPE_HORIZONTAL ||
        currentForegroundMap->mapType == MAPTYPE_VDP)
    {
        ObjectStreamer_HorizontalInit();
        ObjectStreamer_Update = ObjectStreamer_HorizontalUpdate;
    }
    else if (currentForegroundMap->mapType == MAPTYPE_VERTICAL)
    {
        ObjectStreamer_VerticalInit();
        ObjectStreamer_Update = ObjectStreamer_VerticalUpdate;
    }
    else
    {
        ErrorMessage("ObjectStreamer 8way map not supported");
    }
}

void SpawnSection(SpawnList spawnList)
{
    if (spawnList == NULL)
    {
        return;
    }

    while (*spawnList != NULL)
    {
        const SpawnStruct* spawnStruct = *spawnList;

        if (spawnStruct != NULL)
        {
            if (*(streamedObjectStates + spawnStruct->objectId) == UNSPAWNED)
            {
                ObjectStruct* objectStruct = spawnStruct->createObject(spawnStruct->x, spawnStruct->y);
                objectStruct->objectStreamerId = spawnStruct->objectId;

                *(streamedObjectStates + spawnStruct->objectId) = SPAWNED;
            }
        }

        spawnList++;
    }
}

void ObjectStreamer_DestroyObject(ObjectStruct* object)
{
    if (object->objectStreamerId >= 0)
    {
        *(streamedObjectStates + object->objectStreamerId) = UNSPAWNED;
    }
}

void InitSpan(s16 start, s16 end)
{
    SpawnSections sections = currentForegroundMap->spawnSections;
    u16 numSections = currentForegroundMap->numSpawnSections;
    numSections = numSections;

    numberOfSectionsInAScreen = end - start;
    oldCurrentSectionIndex = start;

    int loop = 0;
    for (loop = start; loop <= end; loop++)
    {
        SpawnSection(sections[loop]);
    }
}

void ObjectStreamer_HorizontalInit()
{
    // init the visible screen.
    s16 start = (s16)(VIRTUAL_TO_TILE(foregroundScrollData.scrollx_hires_clamped) / SECTION_SIZE_TILES);
    s16 end = (s16)((VIRTUAL_TO_TILE(foregroundScrollData.scrollx_hires_clamped) + SCREEN_WIDTH_IN_TILES) / SECTION_SIZE_TILES) + 1;

    if (end > currentForegroundMap->numSpawnSections - 1)
    {
        end = currentForegroundMap->numSpawnSections - 1;
    }

    InitSpan(start, end);
}

void ObjectStreamer_VerticalInit()
{
    // init the visible screen.
    s16 start = (s16)(VIRTUAL_TO_TILE(foregroundScrollData.scrolly_hires_clamped) / SECTION_SIZE_TILES);
    s16 end = (s16)((VIRTUAL_TO_TILE(foregroundScrollData.scrolly_hires_clamped) + SCREEN_HEIGHT_IN_TILES) / SECTION_SIZE_TILES) + 1;

    if (end > currentForegroundMap->numSpawnSections - 1)
    {
        end = currentForegroundMap->numSpawnSections - 1;
    }

    InitSpan(start, end);
}



void ObjectStreamer_HorizontalUpdate()
{
    currentSectionIndex = (s16)VIRTUAL_TO_TILE(foregroundScrollData.scrollx_hires_clamped) / SECTION_SIZE_TILES;

    Update();
}

void ObjectStreamer_VerticalUpdate()
{
    currentSectionIndex = (s16)VIRTUAL_TO_TILE(foregroundScrollData.scrolly_hires_clamped) / SECTION_SIZE_TILES;

    Update();
}


void Update()
{
    SpawnSections sections = currentForegroundMap->spawnSections;

    if (currentSectionIndex > oldCurrentSectionIndex) // to the right (or down)
    {
        oldCurrentSectionIndex = currentSectionIndex;

        s16 index = currentSectionIndex + numberOfSectionsInAScreen;

        if (index > currentForegroundMap->numSpawnSections - 1)
        {
            index = currentForegroundMap->numSpawnSections - 1;
        }

        SpawnSection(sections[index]);
    }
    else if (currentSectionIndex < oldCurrentSectionIndex) // to the left (or up)
    {
        oldCurrentSectionIndex = currentSectionIndex;

        SpawnSection(sections[currentSectionIndex]);
    }
}

void PrintSpawnSections()
{
    int loop = 0;
    for (loop = 0; loop < currentForegroundMap->width >> 1; loop++)
    {
        Print("----");
        SpawnList spawnList = currentForegroundMap->spawnSections[loop];

        if (spawnList == NULL)
        {
            Print("Null");
        }
        else
        {
            while (*spawnList != NULL)
            {
                const SpawnStruct* spawnStruct = *spawnList;

                Print("***");

                PrintNumber(spawnStruct->objectId);
                PrintNumber((u32)spawnStruct->createObject);
                PrintNumber(spawnStruct->x);
                PrintNumber(spawnStruct->y);

                spawnList++;
            }
        }
    }
}

