#include "PlayerManager.h"

#include <genesis.h>
#include <kdebug.h>
#include <maths.h>
#include <vdp_spr.h>

#include "generated\AllObjects.h"

#include "GGAnimationDraw.h"
#include "ComplexAnimationDraw.h"
#include "Collisions.h"
#include "Config.h"
#include "Debug.h"
#include "generated\DebugSettings.h"
#include "EnemyTypes.h"
#include "exported\ExportedGenResSprites.h"
#include "exported\ExportedGGAnimations.h"
#include "GameOverScreen.h"
#include "HUDManager.h"
#include "JoypadManager.h"
#include "LevelManager.h"
#include "MathUtils.h"
#include "ObjectManager.h"
#include "ScrollManager.h"
#include "StateManager.h"
#include "TileHeightMaps.h"
#include "VDP_Extra.h"

// game constants.
#define MAX_LIVES                       3
#define MAX_HEALTH_POINTS               4
#define JUMP_SPEED                      29
#define KNOCKBACK_SPEED                 20
#define KNOCKBACK_UP_SPEED              -16
#define DEATH_JUMP_SPEED                17
#define MAX_DOWNATTACK_SPEED            30
#define DOWNATTACK_MIN_FRAMES           7 // how many frames before we can do a down attack.
#define MOVEMENT_SPEED                  10
#define DAMAGE_TIME                     10
#define INVINSIBILITY_TIME              45

// helper macros
#define SET_BODY_ANIM(animation) \
    body.currentVDPInfo = &Penguin##animation##VDPInfo;\
    body.currentAnimation = &Penguin##animation; \
    body.animationFrame = 0;\
    body.animationSpeedCounter = 0;

#define SET_FEET_ANIM(animation) \
    feet.currentVDPInfo = &PenguinFeet##animation##VDPInfo;\
    feet.currentAnimation = &PenguinFeet##animation;\
    feet.animationFrame = 0;\
    feet.animationSpeedCounter = 0;\

#define SET_FEET_ANIM_NONE \
        feet.currentVDPInfo = NULL; \
        feet.currentAnimation = NULL


ObjectStruct player;
ObjectStruct* playerObject;
VDPInfo playerVDPInfo;
PlayerDataStruct playerData;

typedef void (*VoidFunctionType)();

VoidFunctionType updateState = NULL;
VoidFunctionType updatePhysics = NULL;
VoidFunctionType onPhysicsCollision = NULL;
VoidFunctionType postAttackAnimation = NULL;

// penguin body
struct
{
    const VDPInfo* currentVDPInfo;
    const GGAnimation* currentAnimation;
    u16 animationFrame;
    u16 animationSpeed;
    u16 animationSpeedCounter;
} body;

// penguin feet
struct
{
    const VDPInfo* currentVDPInfo;
    const GGAnimation* currentAnimation;
    u16 animationFrame;
    u16 animationSpeedCounter;
} feet;

// various state tracking variables
u16 attacking = 0;
u16 postattack = 0;
u16 deathWaitTime = 0;
s16 jumpCounter = 0;
u16 attackCoolDown = 0;
u16 doubleTap = 0;
u16 damagedTimer = 0;
u16 deathFallCounter = 0;

#define ANIMATION_OFFSET_IDLE       0
#define ANIMATION_OFFSET_RUN        1
#define ANIMATION_OFFSET_JUMP       2
#define ANIMATION_OFFSET_DAMAGED    3
#define ANIMATION_OFFSET_ATTACK     4
#define ANIMATION_OFFSET_DOWNATTACK 5

// y offsets only
#define ANIMATION_OFFSET_IDLE_HIGH      6
#define ANIMATION_OFFSET_RUN_HIGH       7
#define ANIMATION_OFFSET_JUMP_HIGH      8
#define ANIMATION_OFFSET_ATTACK_HIGH    9


s16 animationXOffsetsFacingRight[] =
{
    -39, // idle 
    -39, // run
    -39, // jump
    -15,  // damaged
    -39,  // attack
    -12   // down attack
};

s16 animationXOffsetsFacingLeft[] =
{
    -9, // idle 
    -9, // run
    -9, // jump
    -15,  // damaged
    -57,  // attack
    -12   // down attack
};

s16 animationYOffsets[] =
{
    -8, // idle 
    -2,  // run
    -2,  // jump
    -7,  // damaged
    -16,  // attack
     0,   // down attack
    -8, // idle high
    -10,  // run high
    -10,  // jump high
    -16,  // attack high
};

u16 animationXOffsetIndex = 0;
u16 animationYOffsetIndex = 0;

void StartIdle();
void StartRun();
void StartJump();
void StartFall();
void StartAttack();
void StartDownAttack();
void StartDamaged();
void StartStunned();
void StartDeathFall();

void UpdateIdle();
void UpdateRun();
void UpdateJump();
void UpdateDownAttack();
void UpdateFall();
void UpdateDamaged();
void UpdateDeathFall();
void UpdateStunned();

void UpdateIdlePhysics();
void UpdateRunPhysics();
void UpdateJumpPhysics();
void UpdateFallPhysics();
void UpdateDeathFallPhysics();
void UpdateStunnedPhysics();

void StartIdleAnimation();
void StartRunAnimation();
void StartJumpAnimation();
void StartDamagedAnimation();
void StartAttackAnimation();
void StartAttackUpAnimation();
void StartAttackDownAnimation();
void StartDownAttackAnimation();

void PlayerManager_AfterDeath();


void PlayerManager_Jump()
{
    StartJump();
}

void PlayerManager_Die()
{
    playerObject->healthPoints = 0;

    StartDeathFall();
}

void PlayerManager_CallOnPhysicsCollisionFunction()
{
    if (onPhysicsCollision != NULL)
    {
        onPhysicsCollision();
    }
}

void StartIdle()
{
    playerData.playerState = PLAYERSTATE_IDLE;

    StartIdleAnimation();

    // updates
    updateState = UpdateIdle;
    updatePhysics = UpdateIdlePhysics;
    onPhysicsCollision = NULL;

    // misc
    playerObject->speedx = 0;
    doubleTap = 0;
}

void StartRun()
{
    playerData.playerState = PLAYERSTATE_RUN;

    StartRunAnimation();

    // updates
    updateState = UpdateRun;
    updatePhysics = UpdateRunPhysics;
    onPhysicsCollision = NULL;

    doubleTap = 0;
}


void OnDamagedPhysicsCollision()
{
    updatePhysics = UpdateFallPhysics;
    onPhysicsCollision = NULL;
}

void StartDamaged()
{
    playerData.playerState = PLAYERSTATE_DAMAGED;

    updateState = UpdateDamaged;
    updatePhysics = UpdateJumpPhysics;
    onPhysicsCollision = OnDamagedPhysicsCollision;

    playerObject->speedy = KNOCKBACK_UP_SPEED;

    StartDamagedAnimation();

    playerObject->speedx = -KNOCKBACK_SPEED;

    if (playerObject->direction == DIRECTION_LEFT)
    {
        playerObject->speedx = KNOCKBACK_SPEED;
    }

    doubleTap = 0;
    damagedTimer = DAMAGE_TIME;
    playerData.invincibilityTimer = INVINSIBILITY_TIME;
    playerData.platform = NULL;
}

void StartStunned()
{
    playerData.playerState = PLAYERSTATE_STUNNED;

    updateState = UpdateStunned;
    updatePhysics = UpdateStunnedPhysics;
    onPhysicsCollision = NULL;

    StartDamagedAnimation();

    doubleTap = 0;
    damagedTimer = 60;
}

void PlayerManager_Stun()
{
    StartStunned();
}

void StartJump()
{
    playerData.playerState = PLAYERSTATE_JUMP;

    // body, feet, weapon
    StartJumpAnimation();

    // updates
    updateState = UpdateJump;
    updatePhysics = UpdateJumpPhysics;
    onPhysicsCollision = StartFall;

    // misc
    jumpCounter = 0;
    playerObject->speedy = -JUMP_SPEED;
    doubleTap = 0;
    playerData.platform = NULL;
}

void OnFallPhysicsCollision()
{
    StartIdle();
}

void StartFall()
{
    playerData.playerState = PLAYERSTATE_FALL;

    // body, feet, weapon
    StartJumpAnimation();

    // updates
    updateState = UpdateFall;
    updatePhysics = UpdateFallPhysics;
    onPhysicsCollision = OnFallPhysicsCollision;
    
    // misc
    jumpCounter = 0;
    playerObject->speedy = 0;
    doubleTap = 0;
}

void OnDownAttackPhysicsCollision()
{
    ScrollManager_SetCameraShakeTime(12);
    playerData.freezeTimer = 12;
    playerData.onfreezeTimerDone = StartIdle;
}

void StartDownAttack()
{
    if (playerData.playerState == PLAYERSTATE_JUMP)
    {
        // misc
        jumpCounter = 0;
        playerObject->speedy = 0;
    }

    playerData.playerState = PLAYERSTATE_DOWNATTACK;

    StartDownAttackAnimation();

    // updates
    updateState = UpdateDownAttack;
    updatePhysics = UpdateFallPhysics;
    onPhysicsCollision = OnDownAttackPhysicsCollision;

    doubleTap = 0;
}

void StartDeathFall()
{
    playerData.playerState = PLAYERSTATE_DEAD;

    // updates
    updateState = UpdateDeathFall;
    updatePhysics = UpdateDeathFallPhysics;
    onPhysicsCollision = NULL;

    playerObject->speedy = -DEATH_JUMP_SPEED;
    playerObject->speedx = -MOVEMENT_SPEED;

    if (playerObject->direction == DIRECTION_LEFT)
    {
        playerObject->speedx = MOVEMENT_SPEED;
    }

    deathFallCounter = 180;
    playerData.platform = NULL;

    StartDamagedAnimation();
}

void StartAttack()
{
    if (attacking == 1)
    {
        return;
    }

    StartAttackAnimation();

    attacking = 1;
    doubleTap = 0;
}

void UpdateIdle()
{
    if ((joystate & BUTTON_LEFT) ||
        (joystate & BUTTON_RIGHT))
    {
        StartRun();
    }

    if (joystickState.aPressed)
    {
        // setup the feet when attacking while
        // idle.
        if (attacking != 1)
        {
            SET_FEET_ANIM(Stand);
        }

        StartAttack();
    }

    if (joystickState.bPressed)
    {
        StartJump();
        return;
    }

    if (joystickState.cPressed)
    {
        StartDamaged();
    }
}

void UpdateLeftRight()
{
    playerObject->speedx = 0;

    s16 movementSpeed = MOVEMENT_SPEED;

    if (joystate & BUTTON_LEFT)
    {
        playerObject->speedx = -movementSpeed;
    }
    else if (joystate & BUTTON_RIGHT)
    {
        playerObject->speedx = movementSpeed;
    }

    if ((playerObject->speedx != 0) &&
        (attacking == 0))
    {
        playerObject->direction = (int)(playerObject->speedx < 0);
    }
}



void UpdateRun()
{
    UpdateLeftRight();

    if (((joystickState.rightTapCount == 1) && (joystickState.pressedRightTimer > 0) && (joystickState.pressedRightTimer < 10)) ||
        ((joystickState.leftTapCount == 1) &&  (joystickState.pressedLeftTimer > 0) && (joystickState.pressedLeftTimer < 10)))
    {
        JoypadManager_ResetTaps();
        doubleTap = 1;
    }

    if (joystickState.aPressed)
    {
        StartAttack();
    }

    if (joystickState.bPressed)
    {
        StartJump();
        return;
    }

    if (playerObject->speedx == 0)
    {
        StartIdle();
    }
}

void UpdateDamaged()
{
    playerObject->speedy += 2;

    if (playerObject->speedy > 0)
    {
        updatePhysics = UpdateFallPhysics;
        onPhysicsCollision = NULL;
    }

    playerObject->visible = 0;//(damagedTimer % 2);

    damagedTimer--;

    if (damagedTimer == 0)
    {
        StartFall();
    }
}

void UpdateStunned()
{
    damagedTimer--;

    if (damagedTimer == 0)
    {
        StartIdle();
    }
}

void UpdateJump()
{
    UpdateLeftRight();

    playerObject->speedy += 2;
    jumpCounter++;

    if (playerObject->speedy > 0)
    {
        StartFall();
    }

    if (joystickState.aPressed)
    {
        if ((joystate & BUTTON_DOWN) &&
            (jumpCounter > DOWNATTACK_MIN_FRAMES))
        {
            StartDownAttack();
        }
        else
        {
            StartAttack();
        }
    }
}

void UpdateFall()
{
    UpdateLeftRight();

    playerObject->speedy++;

    if (playerObject->speedy > JUMP_SPEED)
    {
        playerObject->speedy = JUMP_SPEED;
    }

    if (joystickState.aPressed)
    {
        if (joystate & BUTTON_DOWN)
        {
            StartDownAttack();
        }
        else
        {
            StartAttack();
        }
    }
}

void UpdateDownAttack()
{
    UpdateLeftRight();

    playerObject->speedx >>= 1;
    playerObject->speedy += 2;

    if (playerObject->speedy > MAX_DOWNATTACK_SPEED)
    {
        playerObject->speedy = MAX_DOWNATTACK_SPEED;
    }
}

void UpdateDeathFall()
{
    playerObject->speedy++;
    
    deathFallCounter--;

    if (deathFallCounter == 0)
    {
        playerData.lives--;

        if (playerData.lives > 0)
        {
            HUDManager_UpdateLives();
            PlayerManager_AfterDeath();
            StateManager_ChangeToState(LevelManager_StartLevel);
        }
        else
        {
            StateManager_ChangeToState(GameOverScreen_Start);
        }
    }
}

void StartIdleAnimation()
{
    if (attacking == 0)
    {
        if (playerData.swordStance == PLAYER_SWORDSTANCE_DOWN)
        {
            SET_BODY_ANIM(SwordIdle);
            animationYOffsetIndex = ANIMATION_OFFSET_IDLE;
        }
        else
        {
            SET_BODY_ANIM(SwordIdleHigh);
            animationYOffsetIndex = ANIMATION_OFFSET_IDLE_HIGH;
        }

        animationXOffsetIndex = ANIMATION_OFFSET_IDLE;

        // feet
        SET_FEET_ANIM_NONE;
    }
    else
    {
        SET_FEET_ANIM(Stand);
    }

    postAttackAnimation = StartIdleAnimation;
}

void StartRunAnimation()
{
    if (attacking == 0)
    {
        if (playerData.swordStance == PLAYER_SWORDSTANCE_DOWN)
        {
            SET_BODY_ANIM(SwordRun);
            animationYOffsetIndex = ANIMATION_OFFSET_RUN;
        }
        else
        {
            SET_BODY_ANIM(SwordRunHigh);
            animationYOffsetIndex = ANIMATION_OFFSET_RUN_HIGH;
        }

        animationXOffsetIndex = ANIMATION_OFFSET_RUN;
    }

    // feet
    SET_FEET_ANIM(Run);
    
    postAttackAnimation = StartRunAnimation;
}

void StartJumpAnimation()
{
    if (attacking == 0)
    {
        if (playerData.swordStance == PLAYER_SWORDSTANCE_DOWN)
        {
            SET_BODY_ANIM(SwordJump);
            animationYOffsetIndex = ANIMATION_OFFSET_JUMP;
        }
        else
        {
            SET_BODY_ANIM(SwordJumpHigh);
            animationYOffsetIndex = ANIMATION_OFFSET_JUMP_HIGH;
        }

        animationXOffsetIndex = ANIMATION_OFFSET_JUMP;
    }

    // feet
    SET_FEET_ANIM(Jump);

    postAttackAnimation = StartJumpAnimation;
}

void StartDamagedAnimation()
{
    // body
    SET_BODY_ANIM(Hit);

    animationXOffsetIndex = ANIMATION_OFFSET_DAMAGED;

    // feet
    SET_FEET_ANIM_NONE;
}

void StartAttackAnimation()
{
    // body
    if (playerData.swordStance == PLAYER_SWORDSTANCE_DOWN)
    {
        SET_BODY_ANIM(SwordAttack);
        animationYOffsetIndex = ANIMATION_OFFSET_ATTACK;
    }
    else
    {
        SET_BODY_ANIM(SwordAttackHigh);
        animationYOffsetIndex = ANIMATION_OFFSET_ATTACK_HIGH;
    }

    animationXOffsetIndex = ANIMATION_OFFSET_ATTACK;
}

void StartDownAttackAnimation()
{
    SET_BODY_ANIM(SwordDownAttack);

    animationXOffsetIndex = ANIMATION_OFFSET_DOWNATTACK;

    // feet
    SET_FEET_ANIM_NONE;
}

const s16 touchPointLeftOffset = -PIXEL_TO_VIRTUAL(3);
const s16 touchPointRightOffset = PIXEL_TO_VIRTUAL(3);

void UpdatePlatformCheck()
{
    // if two sensors isn't in the platform's rect, then fall.
    if ((playerObject->x + touchPointRightOffset < playerData.platform->x + playerData.platform->rectLeft) ||
        (playerObject->x + touchPointLeftOffset > playerData.platform->x + playerData.platform->rectRight))
    {
        StartFall();
        playerData.platform = NULL;
    }
}

void UpdateGroundCheck()
{
    if ((GetTileType_Virtual(playerObject->x + touchPointRightOffset, playerObject->y + PIXEL_TO_VIRTUAL(32)) == TILE_NONE) &&
        (GetTileType_Virtual(playerObject->x + touchPointLeftOffset, playerObject->y + PIXEL_TO_VIRTUAL(32)) == TILE_NONE))
    {
        StartFall();
    }
}

void UpdateIdlePhysics()
{
    if (playerData.platform != NULL)
    {
        playerObject->speedx = playerData.platform->speedx;
        playerObject->speedy = playerData.platform->speedy;

        playerObject->x += playerObject->speedx;
        playerObject->y += playerObject->speedy;
    }
    else
    {
        UpdateGroundCheck();
    }
}

s16 UpdateLeftRightMovement(s16 speedx)
{
    s16 playerY = VIRTUAL_TO_PIXEL(playerObject->y);
    s16 endXPosition = VIRTUAL_TO_PIXEL(playerObject->x + speedx);

    if (speedx > 0)
    {
        endXPosition += 4;

        if ((endXPosition > TILE_TO_PIXEL(currentForegroundMap->width - 1)) ||
            (IsTileSolid_Pixel(endXPosition, playerY) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 8) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 16) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 24) != 0))
        {
            endXPosition >>= 3;
            endXPosition <<= 3;
            playerObject->x = PIXEL_TO_VIRTUAL(endXPosition - 4);
            speedx = 0;
        }

    }
    else if (speedx < 0)
    {
        endXPosition -= 4;

        if ((endXPosition < 8) ||
            (IsTileSolid_Pixel(endXPosition, playerY) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 8) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 16) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 24) != 0))
        {
            endXPosition >>= 3;
            endXPosition <<= 3;
            endXPosition += 8;
            playerObject->x = PIXEL_TO_VIRTUAL(endXPosition + 4);
            speedx = 0;
        }

    }

    return speedx;
}

void UpdateLeftRightRunPhysics()
{
    // if we're on a platform, we need to add its speed to ours.
    if (playerData.platform != NULL)
    {
        playerObject->speedx += playerData.platform->speedx;
        playerObject->y += playerData.platform->speedy;
    }


    u16 currenttiletype = GetTileType_Virtual(playerObject->x, playerObject->y + PIXEL_TO_VIRTUAL(31));
    u16 currentTile = GetTile_Virtual(playerObject->x, playerObject->y + PIXEL_TO_VIRTUAL(31));
    u16 isCurrentTileFlipped = IS_FLIPPED_HORIZONTALLY(currentTile);

    // determine the running speed from the slope we're on
    s16 effectiveXSpeed = playerObject->speedx;

    if (currenttiletype & TILE_SLOPE)
    {
        u16 slopeHeightMapIndex = ((currenttiletype & 0xf0) >> 4) - 1;

        if (playerObject->speedx < 0)
        {
            if (isCurrentTileFlipped)
            {
                effectiveXSpeed = -slopeClimbUpSpeed[slopeHeightMapIndex];
            }
            else
            {
                effectiveXSpeed = -slopeClimbDownSpeed[slopeHeightMapIndex];
            }
        }
        else if (playerObject->speedx > 0)
        {
            if (isCurrentTileFlipped)
            {
                effectiveXSpeed = slopeClimbDownSpeed[slopeHeightMapIndex];
            }
            else
            {
                effectiveXSpeed = slopeClimbUpSpeed[slopeHeightMapIndex];
            }
        }
    }

    effectiveXSpeed = UpdateLeftRightMovement(effectiveXSpeed);

    if (effectiveXSpeed == 0)
    {
        return;
    }

    // insert x position processing here.
    playerObject->x += effectiveXSpeed;

    // slope processing
    u16 pixelx = VIRTUAL_TO_PIXEL(playerObject->x);
    u16 sensortilex = VIRTUAL_TO_TILE(playerObject->x);
    u16 sensortiley = VIRTUAL_TO_TILE(playerObject->y + PIXEL_TO_VIRTUAL(31));

    u16 nexttiletype = GetTileType_Virtual(playerObject->x, playerObject->y + PIXEL_TO_VIRTUAL(31));

    s16 heightmapindex = 0;


    if (playerObject->speedx != 0)
    {
        // if we're on a slope
        //      if we're moving up (ie going right on an up/right slope) (up/right == down/left)
        //          - check the tile over in the next tile  [1][2]
        //          - if solid                     [2]
        //              check one tile up       [1][X]
        //              if tile is empty
        //                  handle reaching the top
        //                  return
        //          - if tile is slope
        //              handle going to next slope
        //              return

        if (currenttiletype & TILE_SLOPE)
        {
            if ((nexttiletype & TILE_SOLID))
            {
                sensortiley--;
                nexttiletype = GetTileType_Tile(sensortilex, sensortiley);
            }

            // we're on another slope
            if (nexttiletype & TILE_SLOPE)
            {
                // get the appropriate tile height map
                u16 slopeHeightMapIndex = ((nexttiletype & 0xf0) >> 4) - 1;
                const u16* tileHeightMap = slopeHeightMaps[slopeHeightMapIndex];

                // determine the index to use in the tile height map
                u16 nextTile = GetTile_Tile(sensortilex, sensortiley);
                u16 isNextTileFlipped = IS_FLIPPED_HORIZONTALLY(nextTile);

                if (!isNextTileFlipped)
                {
                    heightmapindex = pixelx & 7;
                }
                else
                {
                    heightmapindex = 7 - (pixelx & 7);
                }

                playerObject->y = TILE_TO_VIRTUAL(sensortiley + 1) - PIXEL_TO_VIRTUAL(32) - PIXEL_TO_VIRTUAL(tileHeightMap[heightmapindex]);
            }
            // we're reached the top
            else if (nexttiletype == TILE_NONE)
            {
                playerObject->y = TILE_TO_VIRTUAL(sensortiley + 1) - PIXEL_TO_VIRTUAL(32);
            }
        }
        //  - check if we're going down
        else if (nexttiletype == TILE_NONE)
        {
            sensortiley++;
            nexttiletype = GetTileType_Tile(sensortilex, sensortiley);

            if (nexttiletype & TILE_SLOPE)
            {
                // get the appropriate tile height map
                u16 slopeHeightMapIndex = ((nexttiletype & 0xf0) >> 4) - 1;
                const u16* tileHeightMap = slopeHeightMaps[slopeHeightMapIndex];

                // determine the index to use in the tile height map
                u16 nextTile = GetTile_Tile(sensortilex, sensortiley);
                u16 isNextTileFlipped = IS_FLIPPED_HORIZONTALLY(nextTile);

                if (!isNextTileFlipped)
                {
                    heightmapindex = pixelx & 7;
                }
                else
                {
                    heightmapindex = 7 - (pixelx & 7);
                }

                playerObject->y = TILE_TO_VIRTUAL(sensortiley + 1) - PIXEL_TO_VIRTUAL(32) - PIXEL_TO_VIRTUAL(tileHeightMap[heightmapindex]);
            }
        }
    }
}

void UpdateLeftRightPhysics()
{
    s16 playerY = VIRTUAL_TO_PIXEL(playerObject->y);
    s16 endXPosition = VIRTUAL_TO_PIXEL(playerObject->x + playerObject->speedx);

    if (playerObject->speedx > 0)
    {
        endXPosition += 4;

        if ((endXPosition > TILE_TO_PIXEL(currentForegroundMap->width - 1)) ||
            (IsTileSolid_Pixel(endXPosition, playerY) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 8) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 16) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 24) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 31) != 0))
        {
            endXPosition >>= 3;
            endXPosition <<= 3;
            playerObject->x = PIXEL_TO_VIRTUAL(endXPosition - 4);
            playerObject->speedx = 0;
        }
    }
    else if (playerObject->speedx < 0)
    {
        endXPosition -= 4;

        if ((endXPosition < 8) ||
            (IsTileSolid_Pixel(endXPosition, playerY) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 8) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 16) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 24) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 31) != 0))
        {
            endXPosition >>= 3;
            endXPosition <<= 3;
            endXPosition += 8;
            playerObject->x = PIXEL_TO_VIRTUAL(endXPosition + 4);
            playerObject->speedx = 0;
        }

    }

    playerObject->x += playerObject->speedx;
}

void UpdateRunPhysics()
{
    UpdateLeftRightRunPhysics();

    if (playerData.platform == NULL)
    {
        UpdateGroundCheck();
    }
    else
    {
        UpdatePlatformCheck();
    }
}

void UpdateJumpPhysics()
{
    s16 startPosition = playerObject->y;

    if (startPosition >= 0)
    {
        u16 endPosition = (u16)(startPosition + playerObject->speedy);

        if ((IsTileSolid_Virtual(playerObject->x + touchPointRightOffset, endPosition) != 0) || 
            (IsTileSolid_Virtual(playerObject->x + touchPointLeftOffset, endPosition) != 0))
        {
            startPosition >>= 5;
            startPosition <<= 5;

            playerObject->y = startPosition;
            playerObject->speedy = 0;

            if (onPhysicsCollision != NULL)
            {
                onPhysicsCollision();
            }
        }
    }

    playerObject->y += playerObject->speedy;

    UpdateLeftRightPhysics();
}

//#define PRINT_DEBUG_FALL_INFO 

void UpdateFallPhysics()
{
    // sensor locations on x axis
    s16 rightSensor = playerObject->x + touchPointRightOffset;
    s16 leftSensor = playerObject->x + touchPointLeftOffset;

    // if the bottom pixel is touching a semi-solid tile, then
    // we don't care about collisions as we've passed through it
    s16 startPosition = playerObject->y + PIXEL_TO_VIRTUAL(31);

    if ((IsTileSemiSolid_Virtual(rightSensor, startPosition) != 0) || 
        (IsTileSemiSolid_Virtual(leftSensor, startPosition) != 0))
    {
        goto do_physics;
    }

    // get the pixel immediately underneath the player
    startPosition = playerObject->y + PIXEL_TO_VIRTUAL(32);

    // only care if we're on the map
    if (startPosition > TILE_TO_VIRTUAL(currentForegroundMap->height))
    {
        goto do_physics;
    }

    s16 endPosition = playerObject->y + PIXEL_TO_VIRTUAL(32) + playerObject->speedy;

    // only care if the end position is on the map;
    if (endPosition > TILE_TO_VIRTUAL(currentForegroundMap->height) ||
        endPosition < 0)
    {
        goto do_physics;
    }

    if ((IsTileSolidOrSemiSolid_Virtual(rightSensor, endPosition) != 0) || 
        (IsTileSolidOrSemiSolid_Virtual(leftSensor, endPosition) != 0))
    {
        // touched a solid tile
        endPosition >>= 5;
        endPosition <<= 5;
        playerObject->y = endPosition - PIXEL_TO_VIRTUAL(32);
        playerObject->speedy = 0;

        if (onPhysicsCollision != NULL)
        {
            onPhysicsCollision();
        }
    }
    else // check if we've collided with a slope.
    {
        u16 rawtile = GetTile_Virtual(playerObject->x, endPosition);
        u16 tile = rawtile & TILEINDEX_BITMASK;
        u16 tileType = ARRAY_ITEM_VALUE(currentForegroundMap->tilesetAttributes, tile);

        if (tileType & TILE_SLOPE) // check if we've touched a slope.
        {
                        
            // get the appropriate tile height map
            s16 pixelx = VIRTUAL_TO_PIXEL(playerObject->x);

            u16 slopeHeightMapIndex = ((tileType & 0xf0) >> 4) - 1;
            const u16* tileHeightMap = slopeHeightMaps[slopeHeightMapIndex];

            // determine the height index to use in the tile height map
            s16 heightmapheightindex = 0;
                    
            if (IS_FLIPPED_HORIZONTALLY(rawtile))
            {
                heightmapheightindex = 7 - (pixelx & 7);
            }
            else
            {
                heightmapheightindex = pixelx & 7;
            }

            s16 heightValue = tileHeightMap[heightmapheightindex];

            // see if we're touching the slope.
            s16 slopetiley = VIRTUAL_TO_PIXEL(endPosition) & 7;
            s16 heightvaluey = (7 - heightValue);

            if (slopetiley > heightvaluey)
            {
                // slope hit.
                playerObject->y = endPosition - PIXEL_TO_VIRTUAL(32);
                playerObject->speedy = 0;
                if (onPhysicsCollision != NULL)
                {
                    onPhysicsCollision();
                }
            }
        }
    }

do_physics:
    playerObject->y += playerObject->speedy;

    UpdateLeftRightPhysics();
}

void UpdateDeathFallPhysics()
{
    // don't update if we've gone past the bottom of the screen.
    if (playerObject->spritedef.posy < 224)
    {
        playerObject->x += playerObject->speedx;
        playerObject->y += playerObject->speedy;
    }
}

void UpdateStunnedPhysics()
{

}

void UpdateAttacking()
{
    if (playerData.playerState == PLAYERSTATE_DOWNATTACK)
    {
        SwordDamageCreate(playerObject->x + PIXEL_TO_VIRTUAL(-3), 
                          playerObject->y + PIXEL_TO_VIRTUAL(32), 
                          8, 
                          8,
                          SWORDDAMAGE_DOWNATTACK);
        return;
    }

    if (attacking != 0)
    {
        if (body.animationFrame == 1 &&
            body.animationSpeedCounter == 0)
        {
            if (playerObject->direction == DIRECTION_RIGHT)
            {
                s16 xOffset = PIXEL_TO_VIRTUAL(1);

                SwordDamageCreate(playerObject->x + xOffset, 
                                  playerObject->y - PIXEL_TO_VIRTUAL(16), 
                                  24, 
                                  56,
                                  SWORDDAMAGE_STANDARD_ATTACK);

                SwordDamageCreate(playerObject->x + xOffset + PIXEL_TO_VIRTUAL(24), 
                                  playerObject->y - PIXEL_TO_VIRTUAL(8), 
                                  16, 
                                  40,
                                  SWORDDAMAGE_STANDARD_ATTACK);

                SwordDamageCreate(playerObject->x + xOffset + PIXEL_TO_VIRTUAL(40), 
                                  playerObject->y,
                                  8, 
                                  24,
                                  SWORDDAMAGE_STANDARD_ATTACK);
            }
            else
            {
                s16 xOffset = -PIXEL_TO_VIRTUAL(25);

                SwordDamageCreate(playerObject->x + xOffset, 
                                  playerObject->y - PIXEL_TO_VIRTUAL(16), 
                                  24, 
                                  56,
                                  SWORDDAMAGE_STANDARD_ATTACK);

                SwordDamageCreate(playerObject->x + xOffset - PIXEL_TO_VIRTUAL(16), 
                                  playerObject->y - PIXEL_TO_VIRTUAL(8), 
                                  16, 
                                  40,
                                  SWORDDAMAGE_STANDARD_ATTACK);

                SwordDamageCreate(playerObject->x + xOffset - PIXEL_TO_VIRTUAL(24), 
                                  playerObject->y,
                                  8, 
                                  24,
                                  SWORDDAMAGE_STANDARD_ATTACK);

            }
        }
    }

    if (attackCoolDown > 0)
    {
        attackCoolDown--;
    }
}

void UpdateAutomated()
{
}

void PlayerManager_ReturnControlStartIdle()
{
    StartIdle();
}

void PlayerManager_ReturnControlStartFall()
{
    StartFall();
    playerObject->speedy = JUMP_SPEED;
}

void UpdateAutomatedPhysics()
{
    playerObject->x += playerObject->speedx;
    playerObject->y += playerObject->speedy;

    if (playerObject->speedy > JUMP_SPEED)
    {
        playerObject->speedy = JUMP_SPEED;
    }
}

void PlayerStartAutomatedRun(s16 speed)
{
    playerData.playerState = PLAYERSTATE_AUTOMATED;

    StartRunAnimation();

    playerObject->speedx = speed;

    // updates
    updateState = UpdateAutomated;
    updatePhysics = UpdateAutomatedPhysics;
}

void PlayerManager_StartAutomatedRunLeft()
{
    PlayerStartAutomatedRun(-MOVEMENT_SPEED);
}

void PlayerManager_StartAutomatedRunRight()
{
    PlayerStartAutomatedRun(MOVEMENT_SPEED);
}

void PlayerManager_StartAutomatedFall()
{
    playerData.playerState = PLAYERSTATE_AUTOMATED;

    StartJumpAnimation();

    playerObject->speedy = JUMP_SPEED;

    // updates
    updateState = UpdateAutomated;
    updatePhysics = UpdateAutomatedPhysics;
}

void UpdateDebugMovement()
{
    playerData.oldPositionX = playerObject->x;
    playerData.oldPositionY = playerObject->y;

    if (joystate & BUTTON_LEFT)
    {
        playerObject->speedx = -debugSettings.debugmovementspeed;
    }
    else if (joystate & BUTTON_RIGHT)
    {
        playerObject->speedx = debugSettings.debugmovementspeed;
    }
    else
    {
        playerObject->speedx = 0;
    }

    if (joystate & BUTTON_UP)
    {
        playerObject->speedy = -debugSettings.debugmovementspeed;
    }
    else if (joystate & BUTTON_DOWN)
    {
        playerObject->speedy = debugSettings.debugmovementspeed;
    }
    else
    {
        playerObject->speedy = 0;
    }

    playerObject->x += playerObject->speedx;
    playerObject->y += playerObject->speedy;
}

void UpdateOnGround()
{
    s16 yPosition = playerObject->y + PIXEL_TO_VIRTUAL(32);

    playerData.onGround = (IsTileSolidOrSemiSolid_Virtual(playerObject->x + touchPointRightOffset, yPosition) != 0) || 
                          (IsTileSolid_Virtual(playerObject->x + touchPointLeftOffset, yPosition) != 0);
}

void UpdateHoleDetection()
{
    if (playerData.playerState != PLAYERSTATE_DEAD)
    {
        if (playerObject->y > TILE_TO_VIRTUAL(currentForegroundMap->height + 8))
        {
            StartDeathFall();
        }
    }
}

//#define SHOW_STATE_CHANGES

#ifdef SHOW_STATE_CHANGES
static u16 oldState = -1;
#endif 

void PlayerUpdate(ObjectStruct* playerObject)
{
#ifdef SHOW_STATE_CHANGES
    if (oldState != playerData.playerState)
    {
        switch (playerData.playerState)
        {
        case PLAYERSTATE_IDLE:
            Print("PLAYERSTATE_IDLE");
            break;
        case PLAYERSTATE_JUMP:
            Print("PLAYERSTATE_JUMP");
            break;
        case PLAYERSTATE_RUN:
            Print("PLAYERSTATE_RUN");
            break;
        case PLAYERSTATE_FALL:
            Print("PLAYERSTATE_FALL");
            break;
        case PLAYERSTATE_DOWNATTACK:
            Print("PLAYERSTATE_DOWNATTACK");
            break;
        case PLAYERSTATE_DAMAGED:
            Print("PLAYERSTATE_DAMAGED");
            break;
        case PLAYERSTATE_DEAD:
            Print("PLAYERSTATE_DEAD");
            break;
        case PLAYERSTATE_AUTOMATED:
            Print("PLAYERSTATE_AUTOMATED");
            break;
        case PLAYERSTATE_STUNNED:
            Print("PLAYERSTATE_STUNNED");
            break;
        }

        oldState = playerData.playerState;
    }
#endif
    u16 disabled = playerData.disablePlayer;

    if (disabled == 1 &&
        playerData.playerState == PLAYERSTATE_AUTOMATED)
    {
        disabled = 0;
    }

    if (disabled == 0)
    {
        if (debugSettings.debugmovement == 1)
        {
            UpdateDebugMovement();
        }
        else
        {
            if (playerData.freezeTimer == 0)
            {
                playerData.oldPositionX = playerObject->x;
                playerData.oldPositionY = playerObject->y;

                updateState();
                updatePhysics();

                UpdateAttacking();
                UpdateOnGround();

                if (playerData.playerState != PLAYERSTATE_AUTOMATED)
                {
                    UpdateHoleDetection();
                }

                if (playerData.invincibilityTimer > 0)
                {
                    playerData.invincibilityTimer--;
                }
            }
            else
            {
                playerData.freezeTimer--;

                if (playerData.freezeTimer == 0)
                {
                    if (playerData.onfreezeTimerDone != NULL)
                    {
                        playerData.onfreezeTimerDone();
                        playerData.onfreezeTimerDone = NULL;
                    }
                }

                // if we're frozen yet we're on a moving platform,
                // force update our position. Right now, this is 
                // for when the player down attacks and lands on a 
                // platform. 
                if (playerData.platform != NULL)
                {
                    playerObject->speedx = playerData.platform->speedx;
                    playerObject->speedy = playerData.platform->speedy;

                    playerObject->x += playerObject->speedx;
                    playerObject->y += playerObject->speedy;
                }
            }
        }
    }
}

void PlayerDestroy(ObjectStruct* playerObject)
{
}

void PlayerCollide(ObjectStruct* playerObject, ObjectStruct* collidedObject)
{
    if (debugSettings.debugmovement == 1)
        return;

    if (playerData.invincibilityTimer > 0)
    {
        return;
    }

    if (((collidedObject->objectType == TYPE_ENEMY) ||
        (collidedObject->objectType == TYPE_ENEMY_PROJECTILE)) &&
        (collidedObject->aiState != ENEMY_AI_STATE_DEAD))
    {
        if (playerObject->healthPoints > 0)
        {
            if (debugSettings.invincible == 0)
            {
                playerObject->healthPoints--;
            }

            if (playerObject->healthPoints == 0)
            {
                StartDeathFall();
            }
            else
            {
                StartDamaged();
            }
        }
    }
}

const GGAnimation* oldAnimation = NULL;
u16 oldAnimationFrame = 0;

void PlayerManager_UpdateVDP()
{
    if ((oldAnimation != body.currentAnimation) ||
        (oldAnimationFrame != body.animationFrame))
    {
        VDPManager_CopyGGAnimationFrameToVDP(body.currentAnimation, body.animationFrame, &playerVDPInfo);
        oldAnimation = body.currentAnimation;
        oldAnimationFrame = body.animationFrame;
    }
}

void UpdateBodyAnimation()
{
    if (playerData.freezeTimer == 0)
    {
        // update penguin animation frame.
        body.animationSpeedCounter++;
        if (body.animationSpeedCounter > body.currentAnimation->frameDelayTime)
        {
            body.animationFrame++;
            body.animationSpeedCounter = 0;
        }

        if (body.animationFrame > body.currentAnimation->numberOfFrames - 1)
        {
            if (attacking == 1)
            {
                attacking = 0;
                playerData.swordStance = !playerData.swordStance;
                postAttackAnimation();
            }

            body.animationFrame = 0;
        }
    }

    s16* animationXOffsets = animationXOffsetsFacingRight;

    if (playerObject->direction == DIRECTION_LEFT)
    {
        animationXOffsets = animationXOffsetsFacingLeft;
    }

    GGAnimation_DrawPlayer(playerObject->spritedef.posx + *(animationXOffsets + animationXOffsetIndex), 
                           playerObject->spritedef.posy + *(animationYOffsets + animationYOffsetIndex), 
                           body.currentAnimation, 
                           &playerVDPInfo, 
                           body.animationFrame,
                           playerObject->direction,
                           PAL0);
}


void UpdateFeetAnimation()
{
    if (feet.currentVDPInfo == NULL)
    {
        return;
    }

    s16 yOffset = *(animationYOffsets + animationYOffsetIndex);
    if (attacking == 1)
    {
        yOffset += 14;
    }
    else if (playerData.swordStance == PLAYER_SWORDSTANCE_UP)
    {
        yOffset += 8;
    }

    GGAnimation_Draw(playerObject->spritedef.posx - (feet.currentAnimation->width >> 1), 
                     playerObject->spritedef.posy + 18 + yOffset, 
                     feet.currentAnimation, 
                     feet.animationFrame,
                     playerObject->direction,
                     PAL0);

    if (playerData.freezeTimer > 0)
    {
        return;
    }

    // update feet animation frame
    feet.animationSpeedCounter++;
    if (feet.animationSpeedCounter > feet.currentAnimation->frameDelayTime)
    {
        feet.animationFrame++;
        feet.animationSpeedCounter = 0;
    }
    
    if (feet.animationFrame > feet.currentAnimation->numberOfFrames - 1)
    {
        feet.animationFrame = 0;
    }
}

void PlayerDraw(ObjectStruct* playerObject)
{
    // update screen position
    playerObject->spritedef.posx = (s16)VIRTUAL_TO_PIXEL(playerObject->x - foregroundScrollData.scrollx_hires_clamped);
    playerObject->spritedef.posy = (s16)VIRTUAL_TO_PIXEL(playerObject->y - foregroundScrollData.scrolly_hires_clamped);


    // debugging
    if (debugSettings.showCollisionBox == 1)
    {
        MyVDP_setSprite(playerObject->spritedef.posx - 4, playerObject->spritedef.posy + 24, SPRITE_SIZE(1,1), 1);
        MyVDP_setSprite(playerObject->spritedef.posx - 4, playerObject->spritedef.posy + 16, SPRITE_SIZE(1,1), 1);
        MyVDP_setSprite(playerObject->spritedef.posx - 4, playerObject->spritedef.posy + 8, SPRITE_SIZE(1,1), 1);
        MyVDP_setSprite(playerObject->spritedef.posx - 4, playerObject->spritedef.posy, SPRITE_SIZE(1,1), 1);
    }

    if (playerData.invincibilityTimer & 1)
    {
        return;
    }

    // update animation parts
    UpdateFeetAnimation();
    UpdateBodyAnimation();
}

void PlayerManager_InitPosition(s16 x, s16 y)
{
    playerData.playerLevelStartX = x;
    playerData.playerLevelStartY = y;
    playerData.oldPositionX = x;
    playerData.oldPositionY = y;
}

void PlayerManager_SetupForGame()
{
    playerObject = &player;
    playerObject->updateFunction    = PlayerUpdate;
    playerObject->destroyFunction   = PlayerDestroy;
    playerObject->drawFunction      = PlayerDraw;
    playerObject->collisionFunction = PlayerCollide;
    playerObject->objectState       = OBJECTSTATE_ACTIVE;

    playerObject->objectType        = TYPE_PLAYER;
    playerObject->spritedef.size    = SPRITE_SIZE(3,4);
    playerObject->damage            = 1;
    playerObject->healthPoints      = MAX_HEALTH_POINTS;

    playerObject->rectLeft          = PIXEL_TO_VIRTUAL(-4);
    playerObject->rectRight         = PIXEL_TO_VIRTUAL(3);
    playerObject->rectTop           = PIXEL_TO_VIRTUAL(0);
    playerObject->rectBottom        = PIXEL_TO_VIRTUAL(31);
     
    playerData.numCoins             = 0;
    playerData.lives                = MAX_LIVES;
    playerData.playerLevelStartX    = 0;
    playerData.playerLevelStartY    = 0;
    playerData.oldPositionX         = 0;
    playerData.oldPositionY         = 0;
}

void PlayerManager_AfterDeath()
{
    playerObject->healthPoints      = MAX_HEALTH_POINTS;
}

void PlayerManager_SetupLife()
{
    playerObject->x                 = playerData.playerLevelStartX;
    playerObject->y                 = playerData.playerLevelStartY;
    playerObject->speedx            = 0;
    playerObject->speedy            = 0;
    playerObject->direction         = DIRECTION_RIGHT;


    playerData.invincibilityTimer   = 0;
    playerData.freezeTimer          = 0;
    playerData.onfreezeTimerDone    = NULL;
    playerData.playerState          = 0;
    playerData.disablePlayer        = 0;
    playerData.platform             = NULL;
    playerData.oldPositionX         = playerData.playerLevelStartX;
    playerData.oldPositionY         = playerData.playerLevelStartY;
    playerData.swordStance          = PLAYER_SWORDSTANCE_DOWN;


    attacking = 0;
    deathWaitTime = 0;
    jumpCounter = 0;
    attackCoolDown = 0;
    doubleTap = 0;
    damagedTimer = 0;

    StartIdle();
}
