#include "VDPPlaneManager.h"

#include "Config.h"
#include "Debug.h"
#include "LevelManager.h"
#include "MathUtils.h"
#include "PlayerManager.h"
#include "ScrollManager.h"
#include "VDP_Extra.h"

#include "VDPPlaneManager_VDPMapFunctions.h"
#include "VDPPlaneManager_HorizontalMapFunctions.h"
#include "VDPPlaneManager_VerticalMapFunctions.h"
#include "VDPPlaneManager_8WayMapFunctions.h"

typedef void (*PlaneSetupFunction)();
typedef void (*PlaneUpdateFunction)();
typedef void (*PlaneUpdateVDPFunction)();

PlaneSetupFunction planeSetupForeground;
PlaneUpdateFunction planeUpdateForeground;
PlaneUpdateVDPFunction planeUpdateVDPForeground;

PlaneSetupFunction planeSetupBackground;
PlaneUpdateFunction planeUpdateBackground;
PlaneUpdateVDPFunction planeUpdateVDPBackground;

// internal data used by the various scrolling functions
u16 vdpWidthMinusOne;
u16 vdpHeightMinusOne;

DMADataStruct* dmaDataStruct;
ScrollDataStruct* scrollData;
const VPMap* map;
u32 plan;
u16 palette;
u16 vdpPlaneWidth;
u16 vdpPlaneHeight;

DMADataStruct foregroundDMAData;
DMADataStruct backgroundDMAData;



#define SETDATA(_map, _scrollData, _dmaDataStruct, _plan, _palette) \
    map = _map;\
    dmaDataStruct = &_dmaDataStruct;\
    scrollData = &_scrollData;\
    plan = _plan;\
    palette = _palette



void SetupFunctionPointers(const VPMap* map, 
                           u16 mapType,
                           PlaneSetupFunction* planeSetup,
                           PlaneUpdateFunction* planeUpdate,
                           PlaneUpdateVDPFunction* planeUpdateVDP)
{
    switch (mapType)
    {
    case MAPTYPE_VDP:
        Print("Map Type VDP");
        *planeSetup = VDPPlaneManager_VDPMap_SetupPlane;
        *planeUpdate = VDPPlaneManager_VDPMap_Update;
        *planeUpdateVDP = VDPPlaneManager_VDPMap_UpdateVDP;
        break;
    case MAPTYPE_HORIZONTAL:
        Print("Map Type Horizontal");
        *planeSetup = VDPPlaneManager_HorizontalMap_SetupPlane;
        *planeUpdate = VDPPlaneManager_HorizontalMap_Update;
        *planeUpdateVDP = VDPPlaneManager_HorizontalMap_UpdateVDP;
        break;
    case MAPTYPE_VERTICAL:
        Print("Map Type Vertical");
        *planeSetup = VDPPlaneManager_VerticalMap_SetupPlane;
        *planeUpdate = VDPPlaneManager_VerticalMap_Update;
        *planeUpdateVDP = VDPPlaneManager_VerticalMap_UpdateVDP;
        break;
    case MAPTYPE_8WAY:
        Print("Map Type 8 Way");
        *planeSetup = VDPPlaneManager_8WayMap_SetupPlane;
        *planeUpdate = VDPPlaneManager_8WayMap_Update;
        *planeUpdateVDP = VDPPlaneManager_8WayMap_UpdateVDP;
        break;
    }
}

void VDPPlaneManager_Init()
{
    vdpPlaneWidth = currentForegroundMap->vdpPlaneWidth;
    vdpPlaneHeight = currentForegroundMap->vdpPlaneHeight;

    Assert((vdpPlaneWidth * vdpPlaneHeight) <= 64 * 64, "VDP Plane size error");

    vdpWidthMinusOne = vdpPlaneWidth - 1;
    vdpHeightMinusOne = vdpPlaneHeight - 1;


    if (currentForegroundMap->mapType == MAPTYPE_HORIZONTAL)
    {
        foregroundDMAData.mapDMADataCopyLength = vdpPlaneHeight << 1;
    }
    else if (currentForegroundMap->mapType == MAPTYPE_VERTICAL)
    {
        foregroundDMAData.mapDMADataCopyLength = vdpPlaneWidth << 1;
    }


    // Right now, we go with the simple case where the background will scroll the same way as the foreground.
    // The only limitation is that if the background isn't a VDP type, then their vdpwidth (or vdpheight) needs
    // to be identical to simplify the DMA copy code.
    //
    // if background fits in vdp, then VDP
    // if background is bigger than vdp on both axis, then 8way
    // if background is longer than vdp then Horizontal
    // if background is taller than vdp then Vertical
    //
    // vdp values for background scrolling use foreground vdp values

    u16 backgroundMapType = currentBackgroundMap->mapType;

    if ((currentBackgroundMap->width <= vdpPlaneWidth) && (currentBackgroundMap->height <= vdpPlaneHeight))
    {
        Print("Background fits in VDP");
        backgroundMapType = MAPTYPE_VDP;
    }
    else if ((currentBackgroundMap->width > vdpPlaneWidth) && (currentBackgroundMap->height > vdpPlaneHeight))
    {
        Print("Background is 8way");
        backgroundMapType = MAPTYPE_8WAY;
    }
    else if (currentBackgroundMap->width > vdpPlaneWidth)
    {
        Print("Background is horizontal");
        backgroundMapType = MAPTYPE_HORIZONTAL;
        backgroundDMAData.mapDMADataCopyLength = MIN(currentBackgroundMap->height, vdpPlaneHeight) << 1;
    }
    else if (currentBackgroundMap->height > vdpPlaneHeight)
    {
        Print("Background is vertical");
        backgroundMapType = MAPTYPE_VERTICAL;
        backgroundDMAData.mapDMADataCopyLength = MIN(currentBackgroundMap->width, vdpPlaneWidth) << 1;
    }

    SetupFunctionPointers(currentForegroundMap, currentForegroundMap->mapType, &planeSetupForeground, &planeUpdateForeground, &planeUpdateVDPForeground);
    SetupFunctionPointers(currentBackgroundMap, backgroundMapType, &planeSetupBackground, &planeUpdateBackground, &planeUpdateVDPBackground);

}

void VDPPlaneManager_Update()
{
    SETDATA(currentForegroundMap, foregroundScrollData, foregroundDMAData, APLAN, PAL2);
    planeUpdateForeground();

    SETDATA(currentBackgroundMap, backgroundScrollData, backgroundDMAData, BPLAN, PAL3);
    planeUpdateBackground();
}

void VDPPlaneManager_UpdateVDP()
{
    SETDATA(currentForegroundMap, foregroundScrollData, foregroundDMAData, APLAN, PAL2);
    planeUpdateVDPForeground();

    SETDATA(currentBackgroundMap, backgroundScrollData, backgroundDMAData, BPLAN, PAL3);
    planeUpdateVDPBackground();

}

void VDPPlaneManager_SetupMaps()
{
    VDP_setPlanSize(vdpPlaneWidth, vdpPlaneHeight);

    SETDATA(currentForegroundMap, foregroundScrollData, foregroundDMAData, APLAN, PAL2);
    planeSetupForeground();

    SETDATA(currentBackgroundMap, backgroundScrollData, backgroundDMAData, BPLAN, PAL3);
    planeSetupBackground();
}


