#include "EnemySpider.h"

#include "ObjectCommon.h"

void EnemySpiderUpdate(ObjectStruct* enemyObject);
void EnemySpiderDestroy(ObjectStruct* enemyObject);
void EnemySpiderDraw(ObjectStruct* enemyObject);
void EnemySpiderCollide(ObjectStruct* enemyObject, ObjectStruct* collidedObject);

#define MOVEMENT_SPEED 4

#define ENEMY_AI_STATE_WALK     1
#define ENEMY_AI_STATE_SHOOT    2

#define WALK_TIME       120
#define SHOOT_TIME      50
#define SHOOT_FRAME     40

ObjectStruct* EnemySpiderCreate(s16 x, s16 y)
{
    ObjectStruct* object = ObjectManager_CreateObject(TYPE_ENEMY,
                                                      EnemySpiderUpdate, 
                                                      EnemySpiderDestroy, 
                                                      EnemySpiderDraw,
                                                      EnemySpiderCollide);

    if (!object)
    {
#ifdef DEBUG_OBJECTS
        KDebug_Alert("enemy could not be created!");
#endif
        return NULL;
    }

    object->x                           = x;
    object->y                           = y;
    object->speedx                      = MOVEMENT_SPEED;
    object->speedy                      = 0;
    object->animation                   = &EnemySpider;
    object->aiState                     = ENEMY_AI_STATE_WALK;
    object->rectLeft                    = PIXEL_TO_VIRTUAL(8);
    object->rectRight                   = PIXEL_TO_VIRTUAL(31);
    object->rectTop                     = PIXEL_TO_VIRTUAL(8);
    object->rectBottom                  = PIXEL_TO_VIRTUAL(EnemySpider.height);
    object->healthPoints                = 1;
    object->damage                      = 1;
    object->visible                     = 1;
    object->animationFrame              = 0;
    object->animationSubFrameCounter    = 0;
    object->invulnerableTimer           = 0;
    object->direction                   = 0;
    object->palette                     = PAL1;
    ComputeStandardActiveArea(object);

    if (object->x < playerObject->x)
    {
        object->speedx = MOVEMENT_SPEED;
    }
    else
    {
        object->speedx = -MOVEMENT_SPEED;
    }

#ifdef DEBUG_OBJECTS
    KDebug_Alert("enemyCreate");
#endif
    
    return object;
}

void EnemySpiderUpdateLeftRightPhysics(ObjectStruct* object)
{
    if (object->speedx > 0)
    {
        // if there's no tile at the foot of the enemy, turn around.
        if (IsTileSolid_Virtual(object->x + object->rectRight, object->y + PIXEL_TO_VIRTUAL(EnemySpider.height)) == 0)
        {
            object->speedx = -object->speedx;
            goto updateobjectx;
        }

        s16 startXPosition = object->x + (object->rectRight + PIXEL_TO_VIRTUAL(1));
        s16 endXPosition = startXPosition + object->speedx;

        if (IsTileSolid_Virtual(endXPosition, object->y + PIXEL_TO_VIRTUAL(EnemySpider.height - 1)) != 0)
        {
            endXPosition >>= 5;
            endXPosition <<= 5;

            object->x = endXPosition - (object->rectRight  + PIXEL_TO_VIRTUAL(1));
            object->speedx = -object->speedx;
            goto updateobjectx;
        }


    }
    else if (object->speedx < 0)
    {
        // if there's no tile at the foot of the enemy, turn around.
        if (IsTileSolid_Virtual(object->x + object->rectLeft, object->y + PIXEL_TO_VIRTUAL(EnemySpider.height)) == 0)
        {
            object->speedx = -object->speedx;
            goto updateobjectx;
        }

        s16 startXPosition = object->x + (object->rectLeft - PIXEL_TO_VIRTUAL(1));
        s16 endXPosition = startXPosition + object->speedx;

        if (IsTileSolid_Virtual(endXPosition, object->y + PIXEL_TO_VIRTUAL(EnemySpider.height - 1)) != 0)
        {
            startXPosition >>= 5;
            startXPosition <<= 5;

            object->x = startXPosition - (object->rectLeft - PIXEL_TO_VIRTUAL(1));
            object->speedx = -object->speedx;
            goto updateobjectx;
        }

    }

updateobjectx:
    object->x += object->speedx;
}

void EnemySpiderUpdatePhysics(ObjectStruct* object)
{
    object->speedy = 20;

    s16 startYPosition = object->y + PIXEL_TO_VIRTUAL(object->animation->height);

    if (startYPosition < 0)
    {
        startYPosition = 0;
    }

    s16 endYPosition = startYPosition + object->speedy;

    if ((IsTileSolid_Virtual(object->x + object->rectRight, endYPosition) != 0) ||
        (IsTileSolid_Virtual(object->x + object->rectLeft, endYPosition) != 0))
    {
        endYPosition >>= 5;
        endYPosition <<= 5;
        object->y = endYPosition - PIXEL_TO_VIRTUAL(object->animation->height);
        object->speedy = 0;
    }

    object->y += object->speedy;
}



void EnemySpiderUpdate(ObjectStruct* object)
{
    if (object->aiState == ENEMY_AI_STATE_WALK)
    {
        if (object->invulnerableTimer == 0)
        {
            EnemySpiderUpdatePhysics(object);
            EnemySpiderUpdateLeftRightPhysics(object);
        }

        if (object->lifetime > WALK_TIME)
        {
            if (object->visible)
            {
                object->aiState = ENEMY_AI_STATE_SHOOT;
            }

            object->lifetime = 0;
        }
    }
    else if (object->aiState == ENEMY_AI_STATE_SHOOT)
    {
        if (object->lifetime == SHOOT_FRAME)
        {
            EnemyBulletCreate(object->x + PIXEL_TO_VIRTUAL((object->animation->width >> 1) - 4), object->y, object->speedx << 1, 0);
        }

        if (object->lifetime > SHOOT_TIME)
        {
            object->aiState = ENEMY_AI_STATE_WALK;
            object->lifetime = 0;
        }
    }
    else if (object->aiState == ENEMY_AI_STATE_DEAD)
    {
        object->x += object->speedx;
        object->y += object->speedy >> 2;

        object->speedy++;

        if (object->visible)
        {
            if ((object->lifetime & 7) == 0)
            {
                ExplosionCreate(object->x, object->y);
            }
        }
        else
        {
            object->objectState = OBJECTSTATE_INACTIVE;
        }
    }

    object->spritedef.posx = VIRTUAL_TO_PIXEL(object->x) - foregroundScrollData.scrollx_vdp;
    object->spritedef.posy = VIRTUAL_TO_PIXEL(object->y) - foregroundScrollData.scrolly_vdp;

    if (object->invulnerableTimer > 0)
    {
        object->invulnerableTimer--;
    }
}

void EnemySpiderDestroy(ObjectStruct* object)
{
#ifdef DEBUG_OBJECTS
    KDebug_Alert("EnemySpiderDestroy");
#endif
}

void EnemySpiderDraw(ObjectStruct* object)
{
    object->palette = PAL1;

    if (object->invulnerableTimer > 0)
    {
        object->palette = PAL0;
    }

    object->direction = (object->speedx < 0);

    GGAnimation_DrawObject(object);
    
    u16 headFrame = 0;

    if (object->aiState == ENEMY_AI_STATE_SHOOT)
    {
        headFrame = object->lifetime >> 4;
        if (headFrame > 2)
        {
            headFrame = 0;
        }
    }

    GGAnimation_Draw(object->spritedef.posx + 12, 
                   object->spritedef.posy - 9 - (object->animationFrame & 1),
                   &EnemySpiderHead, 
                   headFrame, 
                   object->direction,
                   object->palette);

    if (object->aiState == ENEMY_AI_STATE_WALK)
    {
        object->animationSubFrameCounter++;

        if (object->animationSubFrameCounter > object->animation->frameDelayTime)
        {
            object->animationFrame++;
            object->animationSubFrameCounter = 0;
        }

        if (object->animationFrame > object->animation->numberOfFrames - 1)
        {
            object->animationFrame = 0;
        }

    }

}

void EnemySpiderCollide(ObjectStruct* object, ObjectStruct* collidedObject)
{
    if (object->invulnerableTimer > 0)
    {
        return;
    }

    if (object->aiState == ENEMY_AI_STATE_DEAD)
    {
        return;
    }

    if (collidedObject->objectType != TYPE_PLAYER_PROJECTILE)
    {
        return;
    }

    object->healthPoints -= collidedObject->damage;

    object->invulnerableTimer = 10;

    if (object->healthPoints <= 0)
    {
        ExplosionCreate(object->x, object->y);
        object->aiState   = ENEMY_AI_STATE_DEAD;
        object->speedy    = 8;
        CoinCreate(object->x, object->y);
    }
}
